import { Grid } from "./grid.ts";
import { Mouse } from "./input.ts";
import { Painter } from "./painter.ts";
import { Renderer } from "./renderer.ts";

export class Simulator {
    private grid: Grid;
    private circuit: Circuit;
    private tooltip: Tooltip;
    private toolbar: Toolbar;

    constructor(
        private mouse: Mouse,
    ) {
        this.grid = new Grid(this.mouse);
        this.circuit = new Circuit(this.grid.transformingMouse());
        this.tooltip = new Tooltip(this.circuit, this.grid.transformingMouse());
        this.toolbar = new Toolbar(this.mouse, this.tooltip);
    }

    render(r: Renderer) {
        hover: {
            if (this.toolbar.hover(this.mouse.x, this.mouse.y) === "break") {
                break hover;
            }
        }

        r.clear("black");
        this.grid.render(r, (tr) => {
            this.circuit.render(tr);
            this.toolbar.render(r);
            this.tooltip.render(tr);
        });
    }
}

class Toolbar {
    private tools: ComponentFactory[] = [
        new SwitchComponent(),
        new LedComponent(),
    ];
    private previews: Component[] = [];

    private lastWidth = 0;
    private lastHeight = 0;

    private hoveringComponentIdx?: number;

    constructor(
        private mouse: Mouse,
        private tooltip: Tooltip,
    ) {
        this.fillPreviews();

        this.mouse.addOnPress(() => {
            const { x, y } = this.mouse;
            for (const [i, component] of this.previews.entries()) {
                if (
                    x >= i * 128 + 96 - 48 &&
                    y >= this.lastHeight - 100 - 48 &&
                    x < i * 128 + 96 + component.width * 32 + 32 &&
                    y < this.lastHeight - 100 + component.height * 32 + 32
                ) {
                    this.tooltip.select(this.tools[i].newInstance());
                    return "stop";
                }
            }
            return "bubble";
        });
    }

    private fillPreviews(): void {
        // this.previews = this.tools.map((tool) => tool.newInstance());
        this.previews = this.tools as unknown as Component[];
    }

    hover(x: number, y: number): "continue" | "break" {
        for (const [i, component] of this.previews.entries()) {
            if (
                x >= i * 128 + 96 - 48 &&
                y >= this.lastHeight - 100 - 48 &&
                x < i * 128 + 96 + component.width * 32 + 32 &&
                y < this.lastHeight - 100 + component.height * 32 + 32
            ) {
                this.hoveringComponentIdx = i;
                return "break";
            }
        }
        this.hoveringComponentIdx = undefined;
        return "continue";
    }

    render(r: Renderer): void {
        this.lastWidth = r.width;
        this.lastHeight = r.height;

        r.fillRect(0, r.height - 200, r.width, 200, "#aaa");
        for (const [i, component] of this.previews.entries()) {
            if (i === this.hoveringComponentIdx) {
                r.fillRect(
                    i * 128 + 96 - 48,
                    r.height - 100 - 48,
                    component.height * 32 + 64,
                    component.width * 32 + 64,
                    "#00000088",
                );
            }
            component.render(r, i * 128 + 96, r.height - 100);
        }
    }
}

class Tooltip {
    private selectedComponent?: Component;

    constructor(
        private circuit: Circuit,
        public mouse: Mouse,
    ) {
        this.mouse.addOnPress(() => {
            if (!this.selectedComponent) {
                return "bubble";
            }

            const x = Math.floor(this.mouse.x / 32);
            const y = Math.floor(this.mouse.y / 32);
            this.circuit.place(this.selectedComponent, x, y);
            this.selectedComponent = undefined;
            return "stop";
        });
    }

    render(r: Renderer): void {
        this.selectedComponent?.render(r, this.mouse.x, this.mouse.y);
    }

    select(component: Component): void {
        this.selectedComponent = component;
    }
}

type PlacedComponent = {
    component: Component;
    x: number;
    y: number;
};

class Circuit {
    private components: PlacedComponent[] = [];

    constructor(
        private mouse: Mouse,
    ) {
        this.mouse.addOnPress(() => {
            for (const { component, x, y } of this.components) {
                if (!component.click) {
                    continue;
                }
                if (
                    this.mouse.x >= x * 32 &&
                    this.mouse.y >= y * 32 &&
                    this.mouse.x < x * 32 + component.width * 32 &&
                    this.mouse.y < y * 32 + component.height * 32
                ) {
                    component.click(
                        this.mouse.x - x * 32,
                        this.mouse.y - y * 32,
                    );
                    return "stop";
                }
            }
            return "bubble";
        });
    }

    place(component: Component, x: number, y: number): void {
        this.components.push({ component, x, y });
    }

    render(r: Renderer): void {
        for (const { component, x, y } of this.components) {
            component.render(r, x * 32 + 16, y * 32 + 16);
        }
    }
}

interface Component {
    get width(): number;
    get height(): number;
    render(r: Renderer, x: number, y: number): void;
    click?(x: number, y: number): void;
}

interface ComponentFactory {
    newInstance(): Component;
}

class SwitchComponent implements Component, ComponentFactory {
    private graphicOn = new Painter(96, 64);
    private graphicOff = new Painter(96, 64);

    private switchOn = false;

    public width = 1;
    public height = 1;

    constructor() {
        this.graphicOff.fillRect(64, 30, 16, 4, "black");
        this.graphicOff.fillCirc(48 + 32, 32, 8, "black");
        this.graphicOff.fillCirc(48, 32, 16, "black");

        this.graphicOn.fillRect(64, 30, 16, 4, "black");
        this.graphicOn.fillCirc(48 + 32, 32, 8, "black");
        this.graphicOn.fillCirc(48, 32, 16, "red");
    }

    newInstance(): Component {
        return new SwitchComponent();
    }

    render(r: Renderer, x: number, y: number): void {
        const graphic = this.switchOn ? this.graphicOn : this.graphicOff;
        graphic.render(r, x - 48, y - 32);
    }

    click(x: number, y: number): void {
        const onButton = Math.sqrt(
            (x - 16) ** 2 + (y - 16) ** 2,
        ) <= 16;

        if (onButton) {
            this.switchOn = !this.switchOn;
        }
    }
}

class LedComponent implements Component, ComponentFactory {
    private graphicOn = new Painter(96, 64);
    private graphicOff = new Painter(96, 64);

    private switchOn = false;

    public width = 1;
    public height = 1;

    constructor() {
        this.graphicOff.fillRect(16, 30, 16, 4, "black");
        this.graphicOff.fillCirc(16, 32, 8, "black");
        this.graphicOff.fillCirc(48, 32, 16, "black");

        this.graphicOn.fillRect(16, 30, 16, 4, "black");
        this.graphicOn.fillCirc(16, 32, 8, "black");
        this.graphicOn.fillCirc(48, 32, 16, "red");
    }

    newInstance(): Component {
        return new LedComponent();
    }

    render(r: Renderer, x: number, y: number): void {
        const graphic = this.switchOn ? this.graphicOn : this.graphicOff;
        graphic.render(r, x - 48, y - 32);
    }
}
